from mutagen.mp4 import MP4, MP4Tags

from tagfiles._common import pack_list, unpack_first
from tagfiles.errors import NonIntegerTag

from .base import BaseTag


class AACTag(BaseTag):

    extensions = ('.m4a', '.m4b')
    mutagen_type = MP4
    tags_type = MP4Tags

    tags_title = ['\xa9nam']
    tags_version = ['----:com.apple.iTunes:VERSION']
    tags_date = ['\xa9day']
    tags_track_number = ['trkn']
    tags_disc_number = ['disk']
    tags_album = ['\xa9alb']
    tags_genre = ['\xa9gen']
    tags_label = ['----:com.apple.iTunes:LABEL']
    tags_catalog_number = ['----:com.apple.iTunes:CATALOGNUMBER']
    tags_release_type = ['----:com.apple.iTunes:RELEASETYPE']
    tags_comment = ['\xa9cmt']

    tags_artist_album = ['aART']
    tags_artist_main = ['\xa9ART']
    tags_artist_remixer = ['----:com.apple.iTunes:REMIXER']
    tags_artist_producer = ['----:com.apple.iTunes:PRODUCER']
    tags_artist_composer = ['\xa9wrt']
    tags_artist_conductor = ['----:com.apple.iTunes:CONDUCTOR']
    tags_artist_djmixer = ['----:com.apple.iTunes:DJMIXER']

    @property
    def version(self):
        return self.get_tag(self.tags_version)

    @version.setter
    def version(self, value):
        self.set_tag(
            fields=self.tags_version,
            value=[unpack_first(value).encode('utf-8')],
            cast_to_str=False,
        )

    @property
    def track_number(self):
        return str(get_num(self.get_tag(self.tags_track_number)))

    @track_number.setter
    def track_number(self, value):
        self.set_list(
            fields=self.tags_track_number,
            values=[
                change_tag_num(self.get_tag(self.tags_track_number), value)
            ],
            cast_to_str=False,
        )

    @property
    def track_total(self):
        return str(get_total(self.get_tag(self.tags_track_number)))

    @track_total.setter
    def track_total(self, value):
        self.set_list(
            fields=self.tags_track_number,
            values=[
                change_tag_total(self.get_tag(self.tags_track_number), value)
            ],
            cast_to_str=False,
        )

    @property
    def disc_number(self):
        return str(get_num(self.get_tag(self.tags_disc_number)))

    @disc_number.setter
    def disc_number(self, value):
        self.set_list(
            fields=self.tags_disc_number,
            values=[
                change_tag_num(self.get_tag(self.tags_disc_number), value)
            ],
            cast_to_str=False,
        )

    @property
    def disc_total(self):
        return str(get_total(self.get_tag(self.tags_disc_number)))

    @disc_total.setter
    def disc_total(self, value):
        self.set_list(
            fields=self.tags_disc_number,
            values=[
                change_tag_total(self.get_tag(self.tags_disc_number), value)
            ],
            cast_to_str=False,
        )

    @property
    def label(self):
        return self.get_tag(self.tags_label)

    @label.setter
    def label(self, value):
        self.set_tag(
            fields=self.tags_label,
            value=unpack_first(value).encode('utf-8'),
            cast_to_str=False,
        )

    @property
    def catalog_number(self):
        return self.get_tag(self.tags_catalog_number)

    @catalog_number.setter
    def catalog_number(self, value):
        self.set_tag(
            fields=self.tags_catalog_number,
            value=unpack_first(value).encode('utf-8'),
            cast_to_str=False,
        )

    @property
    def release_type(self):
        return self.get_tag(self.tags_release_type)

    @release_type.setter
    def release_type(self, value):
        self.set_tag(
            fields=self.tags_release_type,
            value=unpack_first(value).encode('utf-8'),
            cast_to_str=False,
        )

    @property
    def artist_remixer(self):
        return self.get_list(self.tags_artist_remixer)

    @artist_remixer.setter
    def artist_remixer(self, values):
        self.set_list(
            fields=self.tags_artist_remixer,
            values=[v.encode('utf-8') for v in pack_list(values)],
            cast_to_str=False,
        )

    @property
    def artist_producer(self):
        return self.get_list(self.tags_artist_producer)

    @artist_producer.setter
    def artist_producer(self, values):
        self.set_list(
            fields=self.tags_artist_producer,
            values=[v.encode('utf-8') for v in pack_list(values)],
            cast_to_str=False,
        )

    @property
    def artist_conductor(self):
        return self.get_list(self.tags_artist_conductor)

    @artist_conductor.setter
    def artist_conductor(self, values):
        self.set_list(
            fields=self.tags_artist_conductor,
            values=[v.encode('utf-8') for v in pack_list(values)],
            cast_to_str=False,
        )

    @property
    def artist_djmixer(self):
        return self.get_list(self.tags_artist_djmixer)

    @artist_djmixer.setter
    def artist_djmixer(self, values):
        self.set_list(
            fields=self.tags_artist_djmixer,
            values=[v.encode('utf-8') for v in pack_list(values)],
            cast_to_str=False,
        )


def get_num(tag):
    try:
        return int(tag[0])
    except (TypeError, IndexError):
        return tag


def get_total(tag):
    try:
        return int(tag[1])
    except (TypeError, IndexError):
        return None


def change_tag_num(current_tag, value):
    try:
        total = get_total(current_tag)
        return (int(unpack_first(value)), total or 0)
    except ValueError:
        raise NonIntegerTag


def change_tag_total(current_tag, value):
    try:
        num = get_num(current_tag)
        return (num, int(unpack_first(value)))
    except ValueError:
        raise NonIntegerTag



import re
from collections import defaultdict
from itertools import chain

from tagfiles._common import ArtistRoles, TagDate, pack_list, unpack_first


class BaseTag:

    extension = None
    mutagen_type = None
    tags_type = None

    tags_title = []
    tags_version = []
    tags_date = []
    tags_track_number = []
    tags_track_total = []
    tags_disc_number = []
    tags_disc_total = []
    tags_album = []
    tags_genre = []
    tags_label = []
    tags_catalog_number = []
    tags_release_type = []
    tags_comment = []

    tags_artist_album = []
    tags_artist_main = []
    tags_artist_remixer = []
    tags_artist_producer = []
    tags_artist_composer = []
    tags_artist_conductor = []
    tags_artist_djmixer = []

    def __init__(self, filepath, mut):
        self.path = filepath
        self.mut = mut
        if self.mut.tags is None:
            self.create_tags_object()
        self.save = self.mut.save

    def create_tags_object(self):
        self.mut.tags = self.tags_type()

    def __repr__(self):
        return f"<{self.__class__.__name__} object of {self.path}>"

    @classmethod
    def get_allowed_extensions(cls):
        if not hasattr(cls, "_allowed_extensions"):
            cls._allowed_extensions = defaultdict(list)
            for subcls in cls.__subclasses__():
                for ext in subcls.extensions:
                    cls._allowed_extensions[ext].append(subcls)
        return cls._allowed_extensions

    @staticmethod
    def _decode_tag(tag):
        if isinstance(tag, bytes):
            return tag.decode("utf-8")
        elif isinstance(tag, tuple):
            return tag
        return str(tag)

    def _split_values(self, values):
        values = list(
            chain.from_iterable(re.split(r" \\\\ | / |; ", v) for v in values)
        )
        return [v for v in values if v]

    def get_tag(self, fields):
        for t in fields:
            if t in self.mut.tags:
                try:
                    value = unpack_first(self.mut.tags[t])
                    return self._decode_tag(value)
                except (KeyError, IndexError):
                    pass

    def set_tag(self, fields, value, cast_to_str=True):
        if isinstance(value, list):
            if cast_to_str:
                value = r" \\ ".join(str(v) for v in value)
            else:
                value = value[0]
        if fields:
            self.mut.tags[fields[0]] = str(value) if cast_to_str else value

    def get_list(self, fields):
        for t in fields:
            if t in self.mut.tags:
                try:
                    return self._split_values(
                        [self._decode_tag(s) for s in self.mut.tags[t]]
                    )
                except KeyError:
                    pass
        return []

    def set_list(self, fields, values, cast_to_str=True):
        if fields:
            self.mut.tags[fields[0]] = [
                str(v) if cast_to_str else v for v in pack_list(values)
            ]

    @property
    def title(self):
        return self.get_tag(self.tags_title)

    @title.setter
    def title(self, value):
        self.set_tag(self.tags_title, value)

    @property
    def version(self):
        return self.get_tag(self.tags_version)

    @version.setter
    def version(self, value):
        self.set_tag(self.tags_version, value)

    @property
    def date(self):
        return TagDate(self.get_tag(self.tags_date))

    @date.setter
    def date(self, value):
        self.set_tag(self.tags_date, value)

    @property
    def track_number(self):
        return self.get_tag(self.tags_track_number)

    @track_number.setter
    def track_number(self, value):
        self.set_tag(self.tags_track_number, value)

    @property
    def track_total(self):
        return self.get_tag(self.tags_track_total)

    @track_total.setter
    def track_total(self, value):
        self.set_tag(self.tags_track_total, value)

    @property
    def disc_number(self):
        return self.get_tag(self.tags_disc_number)

    @disc_number.setter
    def disc_number(self, value):
        self.set_tag(self.tags_disc_number, value)

    @property
    def disc_total(self):
        return self.get_tag(self.tags_disc_total)

    @disc_total.setter
    def disc_total(self, value):
        self.set_tag(self.tags_disc_total, value)

    @property
    def album(self):
        return self.get_tag(self.tags_album)

    @album.setter
    def album(self, value):
        self.set_tag(self.tags_album, value)

    @property
    def genre(self):
        return self.get_list(self.tags_genre)

    @genre.setter
    def genre(self, values):
        self.set_list(self.tags_genre, values)

    @property
    def label(self):
        return self.get_tag(self.tags_label)

    @label.setter
    def label(self, value):
        self.set_tag(self.tags_label, value)

    @property
    def catalog_number(self):
        return self.get_tag(self.tags_catalog_number)

    @catalog_number.setter
    def catalog_number(self, value):
        self.set_tag(self.tags_catalog_number, value)

    @property
    def release_type(self):
        return self.get_tag(self.tags_release_type)

    @release_type.setter
    def release_type(self, value):
        self.set_tag(self.tags_release_type, value)

    @property
    def comment(self):
        return self.get_tag(self.tags_comment)

    @comment.setter
    def comment(self, value):
        self.set_tag(self.tags_comment, value)

    @property
    def artist_album(self):
        return self.get_list(self.tags_artist_album)

    @artist_album.setter
    def artist_album(self, values):
        self.set_list(self.tags_artist_album, values)

    @property
    def artist(self):
        return {
            ArtistRoles.MAIN: self.artist_main,
            ArtistRoles.FEATURE: self.artist_feature,
            ArtistRoles.REMIXER: self.artist_remixer,
            ArtistRoles.PRODUCER: self.artist_producer,
            ArtistRoles.COMPOSER: self.artist_composer,
            ArtistRoles.CONDUCTOR: self.artist_conductor,
            ArtistRoles.DJMIXER: self.artist_djmixer,
        }

    @property
    def artist_main(self):
        return [
            a
            for a in self.get_list(self.tags_artist_main)
            if not a.startswith("feat. ")
        ]

    @artist_main.setter
    def artist_main(self, values):
        self.set_list(
            self.tags_artist_main,
            pack_list(values) + [f"feat. {a}" for a in self.artist_feature],
        )

    @property
    def artist_feature(self):
        return [
            re.sub(r"^feat\. ", "", a)
            for a in self.get_list(self.tags_artist_main)
            if a.startswith("feat. ")
        ]

    @artist_feature.setter
    def artist_feature(self, values):
        self.set_list(
            self.tags_artist_main,
            [f"feat. {v}" for v in pack_list(values)] + self.artist_main,
        )

    @property
    def artist_remixer(self):
        return self.get_list(self.tags_artist_remixer)

    @artist_remixer.setter
    def artist_remixer(self, values):
        self.set_list(self.tags_artist_remixer, values)

    @property
    def artist_producer(self):
        return self.get_list(self.tags_artist_producer)

    @artist_producer.setter
    def artist_producer(self, values):
        self.set_list(self.tags_artist_producer, values)

    @property
    def artist_composer(self):
        return self.get_list(self.tags_artist_composer)

    @artist_composer.setter
    def artist_composer(self, values):
        self.set_list(self.tags_artist_composer, values)

    @property
    def artist_conductor(self):
        return self.get_list(self.tags_artist_conductor)

    @artist_conductor.setter
    def artist_conductor(self, values):
        self.set_list(self.tags_artist_conductor, values)

    @property
    def artist_djmixer(self):
        return self.get_list(self.tags_artist_djmixer)

    @artist_djmixer.setter
    def artist_djmixer(self, values):
        self.set_list(self.tags_artist_djmixer, values)



from mutagen.flac import FLAC, VCFLACDict
from mutagen.oggopus import OggOpus, OggOpusVComment
from mutagen.oggvorbis import OggVCommentDict, OggVorbis

from .base import BaseTag


class VorbisTag:
    tags_title = ['title']
    tags_version = ['version']
    tags_date = ['date', 'year']
    tags_track_number = ['tracknumber']
    tags_track_total = ['tracktotal']
    tags_disc_number = ['discnumber']
    tags_disc_total = ['disctotal']
    tags_album = ['album']
    tags_genre = ['genre']
    tags_label = ['organization', 'label', 'recordlabel']
    tags_catalog_number = ['catalognumber']
    tags_release_type = ['releasetype']
    tags_comment = ['comment']

    tags_artist_album = ['albumartist']
    tags_artist_main = ['artist']
    tags_artist_remixer = ['remixer']
    tags_artist_producer = ['producer']
    tags_artist_composer = ['composer']
    tags_artist_conductor = ['conductor']
    tags_artist_djmixer = ['djmixer']

    def create_tags_object(self):
        self.mut.add_tags()


class FLACTag(VorbisTag, BaseTag):
    extensions = ('.flac',)
    mutagen_type = FLAC
    tags_type = VCFLACDict


class OggVorbisTag(VorbisTag, BaseTag):
    extensions = ('.ogg',)
    mutagen_type = OggVorbis
    tags_type = OggVCommentDict


class OggOpusTag(VorbisTag, BaseTag):
    extensions = ('.ogg',)
    mutagen_type = OggOpus
    tags_type = OggOpusVComment



from mutagen import id3
from mutagen.id3 import ID3, IPLS, TIPL
from mutagen.mp3 import MP3

from tagfiles._common import unpack_first

from .base import BaseTag

# TODO: Handle v3 and v4 differences wrt TIPL and IPLS by checking type.


class MP3Tag(BaseTag):

    extensions = ('.mp3',)
    mutagen_type = MP3
    tags_type = ID3

    tags_title = ['TIT2']
    tags_version = ['TXXX:VERSION']
    tags_date = ['TDRC', 'TYER']
    tags_track_number = ['TRCK']
    tags_disc_number = ['TPOS']
    tags_album = ['TALB']
    tags_genre = ['TCON']
    tags_label = ['TPUB']
    tags_catalog_number = ['TXXX:CATALOGNUMBER']
    tags_release_type = ['TXXX:RELEASETYPE']
    tags_comment = ['COMM', 'COMM::XXX']

    tags_artist_album = ['TPE2']
    tags_artist_main = ['TPE1']
    tags_artist_remixer = ['TPE4']
    tags_artist_producer = 'producer'
    tags_artist_composer = ['TCOM']
    tags_artist_conductor = ['TPE3']
    tags_artist_djmixer = 'DJ-mix'

    def get_tag(self, fields):
        for t in fields:
            if t in self.mut.tags:
                try:
                    return str(self.mut.tags[t].text[0])
                except (KeyError, ValueError):
                    pass

    def set_tag(self, fields, value):
        if fields:
            if isinstance(value, list):
                value = r' \\ '.join(str(v) for v in value)
            try:  # Assume the tag is a special tag with a description first.
                key, desc = fields[0].split(':', 1)
                kwargs = dict(desc=desc, text=value)
                old_frames = [
                    f
                    for f in self.mut.tags.getall(key)
                    if getattr(f, 'desc', None) != desc
                ]
                self.mut.tags.delall(key)
                for f in old_frames:
                    self.mut.tags.add(f)
            except ValueError:  # The tag isn't a special tag.
                key = fields[0]
                kwargs = dict(text=value)
                self.mut.tags.delall(key)
            frame = getattr(id3, key)(**kwargs)
            self.mut.tags.add(frame)

    def get_paired_text_frame(self, role):
        _, _, full_frame = self._get_full_paired_text_frame()
        if not full_frame:
            return []
        return self._split_values(
            [p[1] for p in full_frame.people if p[0].lower() == role.lower()]
        )

    def _get_full_paired_text_frame(self):
        for tag, frame in {'TIPL': TIPL, 'IPLS': IPLS}.items():
            try:
                return tag, frame, self.mut.tags[tag]
            except KeyError:
                pass
        return 'TIPL', TIPL, None

    def set_paired_text_frame(self, role, value):
        if isinstance(value, list):
            value = r' \\ '.join(str(v) for v in value)

        key, frame_type, full_frame = self._get_full_paired_text_frame()
        if full_frame:
            people = [
                p for p in full_frame.people if p[0].lower() != role.lower()
            ]
        else:
            people = []

        frame = frame_type(encoding=3, people=people + [[role, value]])
        self.mut.tags.delall(key)
        self.mut.tags.add(frame)

    def get_list(self, fields):
        for t in fields:
            if t in self.mut.tags:
                try:
                    return self._split_values(
                        [self._decode_tag(s) for s in self.mut.tags[t]]
                    )
                except KeyError:
                    pass
        return []

    def set_list(self, fields, values):
        self.set_tag(fields, values)

    @property
    def track_number(self):
        return get_num(self.get_tag(self.tags_track_number))

    @track_number.setter
    def track_number(self, value):
        self.set_tag(
            fields=self.tags_track_number,
            value=change_tag_num(self.get_tag(self.tags_track_number), value),
        )

    @property
    def track_total(self):
        return get_total(self.get_tag(self.tags_track_number))

    @track_total.setter
    def track_total(self, value):
        self.set_tag(
            fields=self.tags_track_number,
            value=change_tag_total(
                self.get_tag(self.tags_track_number), value
            ),
        )

    @property
    def disc_number(self):
        return get_num(self.get_tag(self.tags_disc_number))

    @disc_number.setter
    def disc_number(self, value):
        self.set_tag(
            fields=self.tags_disc_number,
            value=change_tag_num(self.get_tag(self.tags_disc_number), value),
        )

    @property
    def disc_total(self):
        return get_total(self.get_tag(self.tags_disc_number))

    @disc_total.setter
    def disc_total(self, value):
        self.set_tag(
            fields=self.tags_disc_number,
            value=change_tag_total(self.get_tag(self.tags_disc_number), value),
        )

    @property
    def artist_producer(self):
        return self.get_paired_text_frame(self.tags_artist_producer)

    @artist_producer.setter
    def artist_producer(self, values):
        self.set_paired_text_frame(self.tags_artist_producer, values)

    @property
    def artist_djmixer(self):
        return self.get_paired_text_frame(self.tags_artist_djmixer)

    @artist_djmixer.setter
    def artist_djmixer(self, values):
        self.set_paired_text_frame(self.tags_artist_djmixer, values)


def get_num(tag):
    try:
        return tag.split('/')[0]
    except (IndexError, AttributeError):
        return tag


def get_total(tag):
    try:
        return tag.split('/')[1]
    except (IndexError, AttributeError):
        return None


def change_tag_num(current_tag, value):
    value = unpack_first(value)
    total = get_total(current_tag)
    return f'{value}/{total}' if total else str(value)


def change_tag_total(current_tag, value):
    num = get_num(current_tag)
    return f'{num}/{unpack_first(value)}'
